package thesis.gui;

import java.awt.Component;
import java.awt.FileDialog;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.DefaultComboBoxModel;
import javax.swing.GroupLayout;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JSpinner;
import javax.swing.JTextField;
import javax.swing.SpinnerNumberModel;
import javax.swing.SwingWorker;
import javax.swing.Timer;
import javax.swing.GroupLayout.Alignment;
import javax.swing.LayoutStyle.ComponentPlacement;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import thesis.simulation.DynamicSimulator;
import thesis.simulation.Integrator;
import thesis.simulation.SequentialSimulator;
import thesis.simulation.Simulator;
import thesis.simulation.StaticSimulator;
import thesis.simulation.UniverseModel;
import thesis.simulation.VideoRecorder;
import thesis.tree.Body;

public final class SimulationDialog extends JDialog {
  
  private static final long serialVersionUID = 1L;
  
  private Simulator simulator;
  private VideoRecorder recorder;
  
  private Timer timer;
  private long simulationStart;
  
  private JTextField modelFile;
  
  private SpinnerNumberModel steps;
  private SpinnerNumberModel stepSize;
  private SpinnerNumberModel openingAngle;
  private SpinnerNumberModel softeningFactor;
  
  private JCheckBox recordVideo;
  private JTextField videoFile;
  
  private JComboBox type;
  private SpinnerNumberModel numThreads;
  private SpinnerNumberModel granularity;
  
  private JProgressBar progress;
  private ImagePanel preview;
  private JLabel totalTime;
  private JLabel[] percentages;
  
  private List<Component> options;
  private List<Component> threadsOptions;
  private List<Component> granularityOptions;
  private List<Component> videoOptions;
  
  private JButton startButton;
  private JButton stopButton;
  
  public SimulationDialog(Window owner) {
    this(owner, "");
  }
  
  public SimulationDialog(Window owner, String modelFilename) {
    super(owner, "Simulacija", DEFAULT_MODALITY_TYPE);
    setDefaultCloseOperation(DISPOSE_ON_CLOSE);
    
    timer = new Timer(1000, new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        updateTotalTime();
      }
    });
    timer.setInitialDelay(0);
    timer.setRepeats(true);
    
    options = new ArrayList<Component>();
    videoOptions = new ArrayList<Component>();
    threadsOptions = new ArrayList<Component>();
    granularityOptions = new ArrayList<Component>();
    
    setResizable(false);
    setContentPane(createContentPanel(modelFilename));
    setOptionsEnable(true);
    pack();
    setLocationRelativeTo(owner);
  }
  
  public void updateTotalTime() {
    long time = System.currentTimeMillis() - simulationStart;
    long seconds = (time / 1000) % 60;
    long minutes = time / 60000;
    
    StringBuilder text = new StringBuilder();
    if (minutes != 0) {
      text.append(minutes).append("m : ");
    }
    text.append(seconds).append("s");
    
    totalTime.setText(text.toString());
  }
  
  public void updateProgress() {
    progress.setValue(progress.getValue() + 1);
    if (progress.getValue() == progress.getMaximum()) {
      stopSimulation();
    }
    
    long[] timers = new long[5];
    long total = 0;
    for (int i = 0; i < timers.length; ++i) {
      timers[i] = simulator.getTimer(i);
      total += timers[i];
    }
    
    DecimalFormat format = new DecimalFormat("0.00");
    for (int i = 0; i < timers.length; ++i) {
      percentages[i].setText(format.format(timers[i] * 100.0 / total) + "%");
    }
    
    if (recorder != null) {
      preview.setImage(recorder.getFrame());
    }
  }
  
  private void startSimulation() {
    setOptionsEnable(false);
    toggleStartStopButtons();
    progress.setValue(0);
    progress.setIndeterminate(true);
    preview.setImage(null);
    
    new SwingWorker<Void, Void>() {
      
      private List<Body> bodies;
      private String error;
      
      @Override
      protected Void doInBackground() throws Exception {
        try {
          bodies = UniverseModel.load(modelFile.getText());
        } catch (FileNotFoundException e) {
          error = "Datoteka modela ne postoji.";
        } catch (IOException e) {
          error = "Dogodila se greška tijekom čitanja modela.";
        }
        return null;
      }
      
      @Override
      protected void done() {
        if (error == null) {
          error =  createSimulator(bodies);
        }
        
        progress.setIndeterminate(false);
        if (error != null) {
          Application.showError(error);
          setOptionsEnable(true);
          toggleStartStopButtons();
        }
      }
      
    }.execute();
  }
  
  private String createSimulator(List<Body> bodies) {
    if (recordVideo.isSelected()) {
      try {
        recorder = new VideoRecorder(videoFile.getText());
      } catch (FileNotFoundException e) {
        return "Datoteku nije moguće koristiti za zapisivanje video snimke.";
      }
    } else {
      recorder = null;
    }
    
    progress.setMaximum(steps.getNumber().intValue());
    progress.setStringPainted(true);
    for (JLabel percentage : percentages) {
      percentage.setText("");
    }
    
    if (type.getSelectedIndex() == 0) {
      simulator = new SequentialSimulator(bodies, steps.getNumber().intValue());
    } else if (type.getSelectedIndex() == 1) {
      simulator = new StaticSimulator(bodies, steps.getNumber().intValue(),
          numThreads.getNumber().intValue());
    } else {
      simulator = new DynamicSimulator(bodies, steps.getNumber().intValue(),
          numThreads.getNumber().intValue(),
          granularity.getNumber().floatValue() / 100.0f);
    }
    
    simulator.setIntegrator(new Integrator(
        (float)Math.pow(10, stepSize.getNumber().intValue()),
        openingAngle.getNumber().floatValue(),
        (float)Math.pow(10, softeningFactor.getNumber().intValue())
    ));
    
    simulator.setRecorder(recorder);
    simulator.setChangeListener(new ChangeListener() {
      @Override
      public void stateChanged(ChangeEvent e) {
        updateProgress();
      }
    });
    
    simulationStart = System.currentTimeMillis();
    timer.start();
    simulator.start();
    
    return null;
  }
  
  private void stopSimulation() {
    timer.stop();
    simulator.interrupt();
    try {
      simulator.join();
    } catch (InterruptedException e) {
      throw new Error(e);   // cannot happen 
    }
    
    if (recorder != null) {
      recorder.close();
      if (recorder.hasFailed()) {
        Application.showError(
            "Dogodila se greška tijekom zapisivanja video snimke.");
      } 
    }
    
    setOptionsEnable(true);
    toggleStartStopButtons();
    progress.setStringPainted(true);
  }

  private void setOptionsEnable(boolean enabled) {
    for (Component component : options) {
      component.setEnabled(enabled);
    }
    
    for (Component component : threadsOptions) {
      component.setEnabled(enabled && type.getSelectedIndex() != 0);
    }
    
    for (Component component : granularityOptions) {
      component.setEnabled(enabled && type.getSelectedIndex() == 2);
    }
    
    for (Component component : videoOptions) {
      component.setEnabled(enabled && recordVideo.isSelected());
    }
  }
  
  private void toggleStartStopButtons() {
    JButton visible = startButton.isVisible() ? startButton : stopButton;
    JButton hidden = visible == startButton ? stopButton : startButton;
    
    hidden.setVisible(true);
    if (visible.hasFocus()) {
      hidden.requestFocusInWindow();
    }
    visible.setVisible(false);
  }

  private JPanel createContentPanel(String modelFilename) {
    JLabel modelLabel = new JLabel("Model:");
    modelFile = new JTextField(modelFilename);
    JButton modelButton = new JButton("...");
    
    options.add(modelLabel);
    options.add(modelFile);
    options.add(modelButton);
    
    modelButton.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        Application.openFileDialog(FileDialog.LOAD, modelFile);
      }
    });
    
    steps = new SpinnerNumberModel(100, 1, 5000, 10);
    JLabel stepsLabel = new JLabel("Broj koraka:");
    JSpinner stepsSpinner = new JSpinner(steps);
    
    stepSize = new SpinnerNumberModel(-3, -9, -1, 1);
    JLabel stepSizeLabel = new JLabel("Veličina koraka:");
    JSpinner stepSizeSpinner = new JSpinner(stepSize);
    
    openingAngle = new SpinnerNumberModel(0.75, 0.25, 2.0, 0.05);
    JLabel openingAngleLabel = new JLabel("Kut otvaranja:");
    JSpinner openingAngleSpinner = new JSpinner(openingAngle);
    
    softeningFactor = new SpinnerNumberModel(-1, -9, -1, 1);
    JLabel softeningFactorLabel = new JLabel("Faktor gušenja:");
    JSpinner softeningFactorSpinner = new JSpinner(softeningFactor);
    
    options.add(stepsLabel);
    options.add(stepsSpinner);
    options.add(stepSizeLabel);
    options.add(stepSizeSpinner);
    options.add(openingAngleLabel);
    options.add(openingAngleSpinner);
    options.add(softeningFactorLabel);
    options.add(softeningFactorSpinner);
    
    JLabel typeLabel = new JLabel("Način rada:");
    type = new JComboBox(new DefaultComboBoxModel(
        new String[] { "Sekvencijalno",
            "Paralelno (statička raspodjela podataka)",
            "Paralelno (dinamička raspodjela podataka)"
        }));
    
    type.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent arg0) {
        setOptionsEnable(true);
      }
    });
    
    options.add(typeLabel);
    options.add(type);
    
    numThreads = new SpinnerNumberModel(2, 2, 64, 1);
    JLabel numThreadsLabel = new JLabel("Broj dretvi:");
    JSpinner numThreadsSpinner = new JSpinner(numThreads);
    
    threadsOptions.add(numThreadsLabel);
    threadsOptions.add(numThreadsSpinner);
    
    granularity = new SpinnerNumberModel(5, 0, 100, 5);
    JLabel granularityLabel = new JLabel("Zrnatost raspodjele:");
    JSpinner granularitySpinner = new JSpinner(granularity);
    
    granularityOptions.add(granularityLabel);
    granularityOptions.add(granularitySpinner);
    
    recordVideo = new JCheckBox("Spremi video snimku");
    recordVideo.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        setOptionsEnable(true);
      }
    });
    
    JLabel videoLabel = new JLabel("Datoteka:");
    videoFile = new JTextField();
    
    JButton videoButton = new JButton("...");
    videoButton.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        Application.openFileDialog(FileDialog.SAVE, videoFile);
      }
    });
    
    options.add(recordVideo);
    videoOptions.add(videoLabel);
    videoOptions.add(videoFile);
    videoOptions.add(videoButton);
    
    JPanel simulationPanel = createSimulationPanel();
    simulationPanel.setBorder(BorderFactory.createEtchedBorder());
    
    JPanel panel = new JPanel();
    GroupLayout layout = new GroupLayout(panel);
    panel.setLayout(layout);
    
    layout.setHorizontalGroup(layout.createSequentialGroup()
      .addContainerGap()
      .addGroup(layout.createParallelGroup()
        .addGroup(layout.createSequentialGroup()
          .addComponent(modelLabel)
          .addPreferredGap(ComponentPlacement.RELATED)
          .addComponent(modelFile)
          .addPreferredGap(ComponentPlacement.RELATED)
          .addComponent(modelButton)
        )
        .addGroup(layout.createSequentialGroup()
          .addGroup(layout.createParallelGroup(Alignment.TRAILING)
            .addComponent(stepsLabel)
            .addComponent(stepSizeLabel)
          )
          .addPreferredGap(ComponentPlacement.RELATED)
          .addGroup(layout.createParallelGroup()
            .addComponent(stepsSpinner, 70, 70, 70)
            .addComponent(stepSizeSpinner, 70, 70, 70)
          )
          .addPreferredGap(ComponentPlacement.UNRELATED)
          .addGroup(layout.createParallelGroup(Alignment.TRAILING)
            .addComponent(openingAngleLabel)
            .addComponent(softeningFactorLabel)
          )
          .addPreferredGap(ComponentPlacement.RELATED)
          .addGroup(layout.createParallelGroup()
            .addComponent(openingAngleSpinner, 70, 70, 70)
            .addComponent(softeningFactorSpinner, 70, 70, 70)
          )
        )
        .addGroup(layout.createSequentialGroup()
          .addComponent(typeLabel)
          .addPreferredGap(ComponentPlacement.RELATED)
          .addComponent(type)
        )
        .addGroup(layout.createSequentialGroup()
          .addComponent(numThreadsLabel)
          .addPreferredGap(ComponentPlacement.RELATED)
          .addComponent(numThreadsSpinner, 60, 60, 60)
          .addPreferredGap(ComponentPlacement.UNRELATED)
          .addComponent(granularityLabel)
          .addPreferredGap(ComponentPlacement.RELATED)
          .addComponent(granularitySpinner, 60, 60, 60)
        )
        .addComponent(recordVideo)
        .addGroup(layout.createSequentialGroup()
          .addComponent(videoLabel)
          .addPreferredGap(ComponentPlacement.RELATED)
          .addComponent(videoFile)
          .addPreferredGap(ComponentPlacement.RELATED)
          .addComponent(videoButton)
        )
        .addComponent(simulationPanel)
      )
      .addContainerGap()
    );
    
    layout.setVerticalGroup(layout.createSequentialGroup()
      .addContainerGap()
        .addGroup(layout.createParallelGroup(Alignment.BASELINE)
          .addComponent(modelLabel)
          .addComponent(modelFile)
          .addComponent(modelButton)
        )
        .addPreferredGap(ComponentPlacement.UNRELATED)
        .addGroup(layout.createParallelGroup(Alignment.BASELINE)
          .addComponent(stepsLabel)
          .addComponent(stepsSpinner)
          .addComponent(openingAngleLabel)
          .addComponent(openingAngleSpinner)
        )
        .addPreferredGap(ComponentPlacement.RELATED)
        .addGroup(layout.createParallelGroup(Alignment.BASELINE)
          .addComponent(stepSizeLabel)
          .addComponent(stepSizeSpinner)
          .addComponent(softeningFactorLabel)
          .addComponent(softeningFactorSpinner)
        )
        .addPreferredGap(ComponentPlacement.UNRELATED)
        .addGroup(layout.createParallelGroup(Alignment.BASELINE)
          .addComponent(typeLabel)
          .addComponent(type)
        )
        .addPreferredGap(ComponentPlacement.RELATED)
        .addGroup(layout.createParallelGroup(Alignment.BASELINE)
          .addComponent(numThreadsLabel)
          .addComponent(numThreadsSpinner)
          .addComponent(granularityLabel)
          .addComponent(granularitySpinner)
        )
        .addPreferredGap(ComponentPlacement.UNRELATED)
        .addComponent(recordVideo)
        .addPreferredGap(ComponentPlacement.RELATED)
        .addGroup(layout.createParallelGroup(Alignment.BASELINE)
          .addComponent(videoLabel)
          .addComponent(videoFile)
          .addComponent(videoButton)
        )
        .addPreferredGap(ComponentPlacement.UNRELATED)
        .addComponent(simulationPanel)
      .addContainerGap()
    );    
    
    return panel;
  }

  private JPanel createSimulationPanel() {
    startButton = new JButton("Pokreni simulaciju");
    startButton.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        startSimulation();
      }
    });
    
    stopButton = new JButton("Zaustavi simulaciju");
    stopButton.setVisible(false);
    stopButton.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        stopSimulation();
      }
    });
    
    progress = new JProgressBar();
    
    JLabel totalTimeLabel = new JLabel("Trajanje simulacije:");
    JLabel treeConstructionLabel = new JLabel("Gradnja stabla:");
    JLabel treeProcessingLabel = new JLabel("Obrada stabla:");
    JLabel forceCalculationLabel = new JLabel("Računanje sila:");
    JLabel integrationLabel = new JLabel("Integracija sila:");
    JLabel videoRecordingLabel = new JLabel("Snimanje videa:");
    
    totalTime = new JLabel("");
    percentages = new JLabel[5];
    for (int i = 0; i < percentages.length; ++i) {
      percentages[i] = new JLabel("");
    }
    
    preview = new ImagePanel(Universe.WIDTH * 3 / 7, Universe.HEIGHT * 3 / 7);
    
    JPanel panel = new JPanel();
    GroupLayout layout = new GroupLayout(panel);
    panel.setLayout(layout);
    
    layout.setHorizontalGroup(layout.createSequentialGroup()
      .addContainerGap()
      .addGroup(layout.createParallelGroup()
        .addGroup(layout.createSequentialGroup()
          .addComponent(startButton)
          .addComponent(stopButton)
        )
        .addComponent(progress, 430, 430, GroupLayout.DEFAULT_SIZE)
        .addGroup(layout.createSequentialGroup()
          .addGroup(layout.createParallelGroup(Alignment.TRAILING)
            .addComponent(totalTimeLabel)
            .addComponent(treeConstructionLabel)
            .addComponent(treeProcessingLabel)
            .addComponent(forceCalculationLabel)
            .addComponent(integrationLabel)
            .addComponent(videoRecordingLabel)
          )
          .addPreferredGap(ComponentPlacement.RELATED)
          .addGroup(layout.createParallelGroup()
            .addComponent(totalTime)
            .addGroup(layout.createParallelGroup(Alignment.TRAILING)
              .addComponent(percentages[0])
              .addComponent(percentages[1])
              .addComponent(percentages[2])
              .addComponent(percentages[3])
              .addComponent(percentages[4])
            )
          )
          .addGap(16, 16, Integer.MAX_VALUE)
          .addComponent(preview)
        )
      )
      .addContainerGap()
    );
    
    layout.setVerticalGroup(layout.createSequentialGroup()
      .addContainerGap()
        .addGroup(layout.createParallelGroup(Alignment.BASELINE)
          .addComponent(startButton)
          .addComponent(stopButton)
        )
        .addPreferredGap(ComponentPlacement.RELATED)
        .addComponent(progress)
        .addPreferredGap(ComponentPlacement.RELATED)
        .addGroup(layout.createParallelGroup()
          .addGroup(layout.createSequentialGroup()
            .addGroup(layout.createParallelGroup(Alignment.BASELINE)
               .addComponent(totalTimeLabel)
               .addComponent(totalTime)
             )
             .addPreferredGap(ComponentPlacement.UNRELATED)
             .addGroup(layout.createParallelGroup(Alignment.BASELINE)
               .addComponent(treeConstructionLabel)
               .addComponent(percentages[0])
             )
             .addPreferredGap(ComponentPlacement.RELATED)
             .addGroup(layout.createParallelGroup(Alignment.BASELINE)
               .addComponent(treeProcessingLabel)
               .addComponent(percentages[1])
             )
             .addPreferredGap(ComponentPlacement.RELATED)
             .addGroup(layout.createParallelGroup(Alignment.BASELINE)
               .addComponent(forceCalculationLabel)
               .addComponent(percentages[2])
             )
             .addPreferredGap(ComponentPlacement.RELATED)
             .addGroup(layout.createParallelGroup(Alignment.BASELINE)
               .addComponent(integrationLabel)
               .addComponent(percentages[3])
             )
             .addPreferredGap(ComponentPlacement.RELATED)
             .addGroup(layout.createParallelGroup(Alignment.BASELINE)
               .addComponent(videoRecordingLabel)
               .addComponent(percentages[4])
             )
           )
           .addComponent(preview)
        )
      .addContainerGap()
    );
    
    return panel;
  }

}
